// SPDX-License-Identifier: MIT
pragma solidity ^0.8.0;

/*

  Something completely different. From the mind of Estebomb
       Developed by (the often controversial) Aleph0ne

   Send 0.01 ETH minimum and 0.1 ETH maximum to this contract
    and you will be given a pre-sale discounted position of
     one item for every 0.01 ETH you send to this contract.

 */
/**
 * @dev Provides information about the current execution context, including the
 * sender of the transaction and its data. While these are generally available
 * via msg.sender and msg.data, they should not be accessed in such a direct
 * manner, since when dealing with meta-transactions the account sending and
 * paying for execution may not be the actual sender (as far as an application
 * is concerned).
 *
 * This contract is only required for intermediate, library-like contracts.
 */
abstract contract Context {
    function _msgSender() internal view virtual returns (address) {
        return msg.sender;
    }

    function _msgData() internal view virtual returns (bytes calldata) {
        return msg.data;
    }
}

/**
 * @dev Contract module which provides a basic access control mechanism, where
 * there is an account (an owner) that can be granted exclusive access to
 * specific functions.
 *
 * By default, the owner account will be the one that deploys the contract. This
 * can later be changed with {transferOwnership}.
 *
 * This module is used through inheritance. It will make available the modifier
 * `onlyOwner`, which can be applied to your functions to restrict their use to
 * the owner.
 */
abstract contract Ownable is Context {
    address private _owner;

    event OwnershipTransferred(
        address indexed previousOwner,
        address indexed newOwner
    );

    /**
     * @dev Initializes the contract setting the deployer as the initial owner.
     */
    constructor() {
        _transferOwnership(_msgSender());
    }

    /**
     * @dev Throws if called by any account other than the owner.
     */
    modifier onlyOwner() {
        _checkOwner();
        _;
    }

    /**
     * @dev Returns the address of the current owner.
     */
    function owner() public view virtual returns (address) {
        return _owner;
    }

    /**
     * @dev Throws if the sender is not the owner.
     */
    function _checkOwner() internal view virtual {
        require(owner() == _msgSender(), "Ownable: caller is not the owner");
    }

    /**
     * @dev Leaves the contract without owner. It will not be possible to call
     * `onlyOwner` functions anymore. Can only be called by the current owner.
     *
     * NOTE: Renouncing ownership will leave the contract without an owner,
     * thereby removing any functionality that is only available to the owner.
     */
    function renounceOwnership() public virtual onlyOwner {
        _transferOwnership(address(0));
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Can only be called by the current owner.
     */
    function transferOwnership(address newOwner) public virtual onlyOwner {
        require(
            newOwner != address(0),
            "Ownable: new owner is the zero address"
        );
        _transferOwnership(newOwner);
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Internal function without access restriction.
     */
    function _transferOwnership(address newOwner) internal virtual {
        address oldOwner = _owner;
        _owner = newOwner;
        emit OwnershipTransferred(oldOwner, newOwner);
    }
}

contract EthReceiver is Ownable {
    uint256 public amtReceived;
    uint256 internal maxAmount = 100000000000000000000;

    event Received(address, uint);
    event Withdrawn(address, uint);
    event ContributionMade(address indexed contributor, uint256 amount);

    // Mapping and array to track sender's contributions
    mapping(address => uint256) public contributions;
    address[] public contributors;

    // The fallback function allows the contract to accept ETH transfers
    receive() external payable {
        require(msg.value >= 10000000000000000, "Too little ETH");
        require(msg.value <= 500000000000000000, "Too much ETH");
        amtReceived += msg.value;

        // Update the contributions
        if (contributions[msg.sender] == 0) {
            contributors.push(msg.sender); // Add new contributor
        }
        contributions[msg.sender] += msg.value;

        require(amtReceived <= maxAmount, "Cannot send more ETH");
        emit Received(msg.sender, msg.value);
        emit ContributionMade(msg.sender, msg.value);
    }

    // Function to get all contributor addresses
    function getAllContributors() external view returns (address[] memory) {
        return contributors;
    }

    // Function to get all contributions
    function getAllContributions() external view returns (uint256[] memory) {
        uint256 length = contributors.length;
        uint256[] memory amounts = new uint256[](length);

        for (uint256 i = 0; i < length; i++) {
            amounts[i] = contributions[contributors[i]];
        }

        return amounts;
    }

    function withdraw() external onlyOwner {
        uint balance = address(this).balance;
        require(balance > 0, "The contract has no ETH");
        payable(msg.sender).transfer(balance);
        emit Withdrawn(msg.sender, balance);
    }
}