// SPDX-License-Identifier: MIT

pragma solidity ^0.8.0;

import "./OFTCoreV2.sol";
import "./IOFTV2.sol";
import "./ERC165.sol";

abstract contract BaseOFTV2 is OFTCoreV2, ERC165, IOFTV2 {
  constructor(uint8 _sharedDecimals, address _lzEndpoint)
    OFTCoreV2(_sharedDecimals, _lzEndpoint)
  {}

  /**
   *
   * public functions
   *
   */
  function sendFrom(
    address _from,
    uint16 _dstChainId,
    bytes32 _toAddress,
    uint256 _amount,
    LzCallParams calldata _callParams
  ) public payable virtual override {
    _send(
      _from,
      _dstChainId,
      _toAddress,
      _amount,
      _callParams.refundAddress,
      _callParams.zroPaymentAddress,
      _callParams.adapterParams
    );
  }

  function sendAndCall(
    address _from,
    uint16 _dstChainId,
    bytes32 _toAddress,
    uint256 _amount,
    bytes calldata _payload,
    uint64 _dstGasForCall,
    LzCallParams calldata _callParams
  ) public payable virtual override {
    _sendAndCall(
      _from,
      _dstChainId,
      _toAddress,
      _amount,
      _payload,
      _dstGasForCall,
      _callParams.refundAddress,
      _callParams.zroPaymentAddress,
      _callParams.adapterParams
    );
  }

  /**
   *
   * public view functions
   *
   */
  function supportsInterface(bytes4 interfaceId)
    public
    view
    virtual
    override(ERC165, IERC165)
    returns (bool)
  {
    return interfaceId == type(IOFTV2).interfaceId
      || super.supportsInterface(interfaceId);
  }

  function estimateSendFee(
    uint16 _dstChainId,
    bytes32 _toAddress,
    uint256 _amount,
    bool _useZro,
    bytes calldata _adapterParams
  ) public view virtual override returns (uint256 nativeFee, uint256 zroFee) {
    return _estimateSendFee(
      _dstChainId, _toAddress, _amount, _useZro, _adapterParams
    );
  }

  function estimateSendAndCallFee(
    uint16 _dstChainId,
    bytes32 _toAddress,
    uint256 _amount,
    bytes calldata _payload,
    uint64 _dstGasForCall,
    bool _useZro,
    bytes calldata _adapterParams
  ) public view virtual override returns (uint256 nativeFee, uint256 zroFee) {
    return _estimateSendAndCallFee(
      _dstChainId,
      _toAddress,
      _amount,
      _payload,
      _dstGasForCall,
      _useZro,
      _adapterParams
    );
  }

  function circulatingSupply() public view virtual override returns (uint256);

  function token() public view virtual override returns (address);
}
