// SPDX-License-Identifier: None

pragma solidity >=0.8.0 <0.9.0;

import "openzeppelin/contracts/utils/introspection/IERC165.sol";

/**
 * dev Interface for the NFT Royalty Standard
 * author Taken from https://eips.ethereum.org/EIPS/eip-2981
 */
interface IERC2981 is IERC165 {
    /**
     * notice Called with the sale price to determine how much royalty
     * is owed and to whom.
     * param tokenId - the NFT asset queried for royalty information
     * param salePrice - the sale price of the NFT asset specified by _tokenId
     * return receiver - address of who should be sent the royalty payment
     * return royaltyAmount - the royalty payment amount for _salePrice
     */
    function royaltyInfo(uint256 tokenId, uint256 salePrice)
        external
        view
        returns (address receiver, uint256 royaltyAmount);
}
