pragma solidity >=0.8.0 <0.9.0;

// SPDX-License-Identifier: MIT

/**
 * title The interface of an optional AS resolver.
 */
interface IASResolver {
    /**
     * dev Returns whether the resolver supports ETH transfers
     */
    function isPayable() external pure returns (bool);

    /**
     * dev Resolves an attestation and verifier whether its data conforms to the spec.
     *
     * param recipient The recipient of the attestation.
     * param schema The AS data schema.
     * param data The actual attestation data.
     * param expirationTime The expiration time of the attestation.
     * param msgSender The sender of the original attestation message.
     *
     * return Whether the data is valid according to the scheme.
     */
    function resolve(
        address recipient,
        bytes calldata schema,
        bytes calldata data,
        uint256 expirationTime,
        address msgSender
    ) external payable returns (bool);
}
