pragma solidity >=0.8.0 <0.9.0;

// SPDX-License-Identifier: MIT

/**
 * title EIP712 typed signatures verifier for EAS delegated attestations interface.
 */
interface IEASEIP712Verifier {
    /**
     * dev Returns the current nonce per-account.
     *
     * param account The requested accunt.
     *
     * return The current nonce.
     */
    function getNonce(address account) external view returns (uint256);

    /**
     * dev Verifies signed attestation.
     *
     * param recipient The recipient of the attestation.
     * param schema The UUID of the AS.
     * param expirationTime The expiration time of the attestation.
     * param refUUID An optional related attestation's UUID.
     * param data Additional custom data.
     * param attester The attesting account.
     * param v The recovery ID.
     * param r The x-coordinate of the nonce R.
     * param s The signature data.
     */
    function attest(
        address recipient,
        bytes32 schema,
        uint256 expirationTime,
        bytes32 refUUID,
        bytes calldata data,
        address attester,
        uint8 v,
        bytes32 r,
        bytes32 s
    ) external;

    /**
     * dev Verifies signed revocations.
     *
     * param uuid The UUID of the attestation to revoke.
     * param attester The attesting account.
     * param v The recovery ID.
     * param r The x-coordinate of the nonce R.
     * param s The signature data.
     */
    function revoke(
        bytes32 uuid,
        address attester,
        uint8 v,
        bytes32 r,
        bytes32 s
    ) external;
}
