// SPDX-License-Identifier: GPL-3.0

pragma solidity ^0.8.17;

struct ExactInputParams {
    bytes path;
    address recipient;
    uint256 deadline;
    uint256 amountIn;
    uint256 amountOutMinimum;
}

/// title Router token swapping functionality
/// notice Functions for swapping tokens via Uniswap V3
interface IUniswapV3Router {
    /// notice Swaps `amountIn` of one token for as much as possible of another along the specified path
    /// param params The parameters necessary for the multi-hop swap, encoded as `ExactInputParams` in calldata
    /// return amountOut The amount of the received token
    function exactInput(ExactInputParams calldata params) external payable returns (uint256 amountOut);
}

/// title Router for price estimation functionality
/// notice Functions for getting the price of one token with respect to another using Uniswap V2
/// dev This interface is only used for non critical elements of the protocol
interface IUniswapV2Router {
    /// notice Given an input asset amount, returns the maximum output amount of the
    /// other asset (accounting for fees) given reserves.
    /// param path Addresses of the pools used to get prices
    function getAmountsOut(uint256 amountIn, address[] calldata path) external view returns (uint256[] memory amounts);

    function swapExactTokensForTokens(
        uint256 swapAmount,
        uint256 minExpected,
        address[] calldata path,
        address receiver,
        uint256 swapDeadline
    ) external;
}
