// SPDX-License-Identifier: MIT

pragma solidity ^0.8.9;

import "../dependencies/openzeppelin/utils/structs/EnumerableSet.sol";
import "../interfaces/IPoolRegistry.sol";
import "../interfaces/external/IMasterOracle.sol";

abstract contract PoolRegistryStorageV1 is IPoolRegistry {
    /**
     * notice Pools collection
     */
    EnumerableSet.AddressSet internal pools;

    /**
     * notice Prices' oracle
     */
    IMasterOracle public override masterOracle;

    /**
     * notice Fee collector address
     */
    address public override feeCollector;

    /**
     * notice Native token gateway address
     */
    address public override nativeTokenGateway;

    /**
     * notice Map of the ids of the pools
     */
    mapping(address => uint256) public override idOfPool;

    /**
     * notice Counter of ids of the pools
     */
    uint256 public override nextPoolId;
}
